package br.usp.ime.mac499.xba.codegen;

import static org.objectweb.asm.Opcodes.ACC_PRIVATE;
import static org.objectweb.asm.Opcodes.ACC_PUBLIC;
import static org.objectweb.asm.Opcodes.ACC_STATIC;
import static org.objectweb.asm.Opcodes.ALOAD;
import static org.objectweb.asm.Opcodes.ARETURN;
import static org.objectweb.asm.Opcodes.ASTORE;
import static org.objectweb.asm.Opcodes.BIPUSH;
import static org.objectweb.asm.Opcodes.DUP;
import static org.objectweb.asm.Opcodes.GETFIELD;
import static org.objectweb.asm.Opcodes.GETSTATIC;
import static org.objectweb.asm.Opcodes.GOTO;
import static org.objectweb.asm.Opcodes.H_INVOKESTATIC;
import static org.objectweb.asm.Opcodes.ICONST_0;
import static org.objectweb.asm.Opcodes.IF_ACMPEQ;
import static org.objectweb.asm.Opcodes.INVOKESPECIAL;
import static org.objectweb.asm.Opcodes.INVOKESTATIC;
import static org.objectweb.asm.Opcodes.INVOKEVIRTUAL;
import static org.objectweb.asm.Opcodes.NEW;
import static org.objectweb.asm.Opcodes.POP;
import static org.objectweb.asm.Opcodes.PUTFIELD;
import static org.objectweb.asm.Opcodes.PUTSTATIC;
import static org.objectweb.asm.Opcodes.RETURN;
import static org.objectweb.asm.Opcodes.SIPUSH;
import static org.objectweb.asm.Opcodes.SWAP;
import static org.objectweb.asm.Opcodes.V1_7;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.util.Iterator;
import java.util.Stack;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Handle;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Type;

import br.usp.ime.mac499.xba.ast.ASTAssignment;
import br.usp.ime.mac499.xba.ast.ASTBinaryMessage;
import br.usp.ime.mac499.xba.ast.ASTBlock;
import br.usp.ime.mac499.xba.ast.ASTBooleanLiteral;
import br.usp.ime.mac499.xba.ast.ASTClass;
import br.usp.ime.mac499.xba.ast.ASTExportStatement;
import br.usp.ime.mac499.xba.ast.ASTFieldDefinition;
import br.usp.ime.mac499.xba.ast.ASTFunctionApplication;
import br.usp.ime.mac499.xba.ast.ASTIdentifier;
import br.usp.ime.mac499.xba.ast.ASTIf;
import br.usp.ime.mac499.xba.ast.ASTInteger;
import br.usp.ime.mac499.xba.ast.ASTLambda;
import br.usp.ime.mac499.xba.ast.ASTLet;
import br.usp.ime.mac499.xba.ast.ASTMethodDefinition;
import br.usp.ime.mac499.xba.ast.ASTModule;
import br.usp.ime.mac499.xba.ast.ASTModule.InstanceInfo;
import br.usp.ime.mac499.xba.ast.ASTSelf;
import br.usp.ime.mac499.xba.ast.ASTString;
import br.usp.ime.mac499.xba.ast.ASTUnaryMessage;
import br.usp.ime.mac499.xba.ast.Node;
import br.usp.ime.mac499.xba.runtime.XBABool;
import br.usp.ime.mac499.xba.runtime.XBAFunction;
import br.usp.ime.mac499.xba.runtime.XBAInteger;
import br.usp.ime.mac499.xba.runtime.XBAObject;
import br.usp.ime.mac499.xba.runtime.XBAString;
import br.usp.ime.mac499.xba.runtime.XBAUnit;
import br.usp.ime.mac499.xba.types.inference.TypeInferer;
import br.usp.ime.mac499.xba.util.BindingsManager;


/**
 * Gera código de uma classe, tomando como entrada um programa já parciado, isto é, 
 * uma árvore sintática.
 * 
 * Está grande porque é um visitor que tem que lidar com cada tipo de nó da árvore 
 * sintatica (infelizmente java nao tem parcial class.)
 * 
 * FALTA INTEGRAR com a inferência de tipos TypeInferer. Por enquanto a inferência 
 * é feita separadamente em TypeChecker.
 * @author stagni
 *
 */
public class XBACodeGen {
	private static final Handle bootstrap = new Handle(H_INVOKESTATIC,
			  "br/usp/ime/mac499/xba/runtime/XBACallSite", "bootstrap",
			  "(Ljava/lang/invoke/MethodHandles$Lookup;Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/CallSite;");
	
	
	// CURR STATE
	private String className;
	private ClassWriter cw;
	private Stack<MethodVisitor> mvs = new Stack<MethodVisitor>();
	private MethodVisitor mvStatic;
	private BindingsManager<BindedIdentifier> bm = new BindingsManager<BindedIdentifier>();
	private TypeInferer ti = new TypeInferer(bm);
	
	
	private String internalClassName() {
		return className.replaceAll("\\.", "\\/");
	}
	
	private String classTypeDescriptor() {
		return "L" + internalClassName() + ";";
	}
	private int nLocalsVars;
	private void startNewMethod(int access, String name, String desc) {
		MethodVisitor mv = cw.visitMethod(access, name, desc, null, null);
		mv.visitCode();
		mvs.push(mv);
		nLocalsVars = 0;
	}
	
	private void endCurrMethod() {
		currMethod().visitInsn(ARETURN);
		MethodVisitor mv = mvs.pop();
		mv.visitMaxs(-1, -1);
		mv.visitEnd();
	}
	
	private MethodVisitor currMethod() {
		return mvs.peek();
	}
	
	public byte[] generateCode(Node node) {
		cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES);
		process(node);
		return cw.toByteArray();
	}
	
	public br.usp.ime.mac499.xba.types.Type process(ASTModule module) {
		bm.addNewScope();
		for (String className : module.classImports()) {
			String soloClassName = soloClassName(className);	
			bm.addBinding(soloClassName, new ClassId(className, ti.typeOfClassId(className)));
		}
		for (InstanceInfo instanceImport : module.instanceImports()) {
			bm.addBinding(instanceImport.instanceName,
					new DefaultInstanceIdentifier(instanceImport.className, 
							instanceImport.instanceName, null));//FIXME
		}
		processChildren(module);
		bm.exitCurrScope();
		return null;
	}
	
	private String soloClassName(String fullClassName) {
		return fullClassName.substring(fullClassName.lastIndexOf('/') + 1);
	}
	
	public br.usp.ime.mac499.xba.types.Type process(ASTClass classz) {
		cw.visit(V1_7, ACC_PUBLIC, classz.name(), null, iname(XBAObject.class), null);
		className = classz.name();
		initializeStaticConstructor();
		bm.addNewScope();
		bm.addBinding(soloClassName(classz.name()), new ClassId(classz.name(), null));
		
		addFields(classz);		
		addConstructor(classz);
		
		for (ASTMethodDefinition method : classz.methods()) 
			process(method);	
		
		for (ASTExportStatement export : classz.exports())
			processExport(export);
		finalizeStaticConstructor();	
		cw.visitEnd();
		bm.exitCurrScope();
		return null;
	}

	protected void processExport(ASTExportStatement export) {
		// FIXME: Impedir uso de 'self' aqui. Talvez seja melhor não deixar self como 
		// palavra reservada, mas como identificador bindado no início do processamento 
		// da classe.
		String fieldName = "$$" + export.id();
		cw.visitField(ACC_PUBLIC + ACC_STATIC, fieldName, desc(XBAObject.class), null, null).visitEnd();
		mvs.push(mvStatic);
			process(export.inicializedTo());
		mvs.pop();
		mvStatic.visitFieldInsn(PUTSTATIC, internalClassName(),	fieldName, desc(XBAObject.class));
	}
	

	private void addConstructor(ASTClass classz) {
		addDefaultConstructor();
		//if (classz.numberOfParameters() == 0) { // XXX HACK
		//	classz.addSelector(new Token(0, "new"), new Token(0, "$hack"));
		//} 
		String constructorName = classz.methodName();
		String descriptor = classz.selectorsDescriptor(); 
		
		MethodType consType = XBAObject.rraType(classz.numberOfParameters());
		startNewMethod(ACC_PUBLIC + ACC_STATIC, constructorName, consType.toMethodDescriptorString());
		
		// XXX HACK como é estático, o 1o argumento vem na variável 0. Mas vamos trocá-lo pela 
		// última, reservando a variável 0 para o objeto que está sendo construído.
		// === BINDING DOS PARAMETROS ===
		currMethod().visitVarInsn(ALOAD, 0);
		currMethod().visitVarInsn(ASTORE, classz.numberOfParameters());
		bm.addNewScope();
		for (int k = 1; k < classz.numberOfParameters(); k++) {
			bm.addBinding(classz.parameter(k), new LocalIdentifier(k, ti.typeOfMethodArgument(classz, k))); 
		}
		bm.addBinding(classz.parameter(0), new LocalIdentifier(classz.numberOfParameters(), 
				ti.typeOfMethodArgument(classz, 0)));
		
		// 	=== CRIANDO O OBJETO ===
		currMethod().visitTypeInsn(NEW, internalClassName());
		currMethod().visitInsn(DUP);
		currMethod().visitMethodInsn(INVOKESPECIAL, internalClassName(), "<init>", "()V");
		currMethod().visitVarInsn(ASTORE, 0); // Colocando obj na posicao 0
		
		this.nLocalsVars = classz.numberOfParameters() + 1;
		
		// === INICIALIZANDO FIELDS ===
		for (ASTFieldDefinition fieldDef : classz.fields()) {
			process(fieldDef.initializedTo());
			BindedIdentifier fieldVar = bm.lookup(fieldDef.name());
			fieldVar.generateStore(this);
		}
		// === INVOCANDO INIT() CASO EXISTA
		if (classz.hasInitMethod()) {
			currMethod().visitVarInsn(ALOAD, 0);
			currMethod().visitFieldInsn(GETSTATIC, iname(XBAUnit.class), "instance", desc(XBAUnit.class));
			currMethod().visitInvokeDynamicInsn("init",
					XBAObject.rraRawType.toMethodDescriptorString(), bootstrap);
			currMethod().visitInsn(POP);
		} 
		currMethod().visitVarInsn(ALOAD, 0);
		endCurrMethod();
		
		pushXBAFunctionOf(constructorName, descriptor, true);
		cw.visitField(ACC_PUBLIC + ACC_STATIC, "$constructor", desc(XBAFunction.class), null, null).visitEnd();
		mvStatic.visitFieldInsn(PUTSTATIC, internalClassName(),
				"$constructor", desc(XBAFunction.class));
		
	}
	
	private void addDefaultConstructor() {
		MethodVisitor mvCons = cw.visitMethod(ACC_PUBLIC, "<init>", "()V", null, null);
		mvCons.visitCode();
		mvCons.visitVarInsn(ALOAD, 0);
		mvCons.visitMethodInsn(INVOKESPECIAL, "java/lang/Object", "<init>", "()V");
		mvCons.visitInsn(RETURN);
		mvCons.visitMaxs(1, 1);
		mvCons.visitEnd();
	}

	private void finalizeStaticConstructor() {
		mvStatic.visitInsn(RETURN);
		mvStatic.visitMaxs(-1, -1);
		mvStatic.visitEnd();
	}

	private void initializeStaticConstructor() {
		mvStatic = cw.visitMethod(ACC_STATIC, "<clinit>", "()V", null, null);
		mvStatic.visitCode();
		/* Deixa uma instancia de lookup no topo da pilha. Será usado durante a construcao de cada 
		 * método para a introdução de campos estáticos contendo Function de cada um dos métodos
		 */
		mvStatic.visitMethodInsn(INVOKESTATIC, "java/lang/invoke/MethodHandles",
				"lookup", "()Ljava/lang/invoke/MethodHandles$Lookup;");
	}

	private void addFields(ASTClass classz) {
		for (ASTFieldDefinition field : classz.fields()) { 
			String fieldName = field.name();
			cw.visitField(ACC_PRIVATE, "$field$" + fieldName, desc(XBAObject.class), null, null).visitEnd();
			bm.addBinding(fieldName, new FieldVariable("$field$" + fieldName, internalClassName(),
					ti.typeOfField(field)));
		}
	}
	
	public br.usp.ime.mac499.xba.types.Type process(ASTMethodDefinition method) {
		MethodType methodType = XBAObject.rraType(method.numberOfParameters()); 
		String descriptor = methodType.toMethodDescriptorString();
		startNewMethod(ACC_PUBLIC, method.methodName(), descriptor);
		bm.addNewScope();
		for (int k = 0; k < method.numberOfParameters(); k++) {
			bm.addBinding(method.parameter(k), new LocalIdentifier(k+1, ti.typeOfMethodArgument(method, k)));
		}
		this.nLocalsVars = method.numberOfParameters() + 1;
		process(method.body());
		currMethod().visitInsn(ARETURN);
		endCurrMethod();
		bm.exitCurrScope();
		
		// A partir daqui cria campo estático contendo Function do método que foi definido
		cw.visitField(ACC_PUBLIC + ACC_STATIC, method.selector(0), desc(XBAFunction.class), null, null).visitEnd();
		pushXBAFunctionOf(method.methodName(), method.selectorsDescriptor(), false);
		mvStatic.visitFieldInsn(PUTSTATIC, internalClassName(), 
				method.selector(0), desc(XBAFunction.class));
		return null;
	}
	
	public static String iname(Class<?> classz) {
		return Type.getInternalName(classz);
	}
	
	public static String desc(Class<?> classz) {
		return Type.getDescriptor(classz);
	}
	// FIXME: colocar esses auxiliares em classe a parte
	
	public static String methodDesc(Class<?> returnT, Class<?>... args) {
		Type[] argTypes = new Type[args.length];
		for (int k = 0; k < args.length; k++)
			argTypes[k] = Type.getType(args[k]);
		return Type.getMethodDescriptor(Type.getType(returnT), argTypes);
	}
	
	private void pushXBAFunctionOf(String methodName, String selectorDescriptor, boolean isStatic) {
		String make = (isStatic)? "makeXBAFunctionOfStatic" : "makeXBAFunctionOf";
		
		// push class
		mvStatic.visitLdcInsn(Type.getType(classTypeDescriptor()));
		// push methodName
		mvStatic.visitLdcInsn(methodName);		
		// push selectors[]
		mvStatic.visitLdcInsn(selectorDescriptor);
		mvStatic.visitLdcInsn(":");
		mvStatic.visitMethodInsn(INVOKEVIRTUAL, iname(String.class), 
				"split", methodDesc(String[].class, String.class));
		mvStatic.visitMethodInsn(INVOKESTATIC, iname(XBAFunction.class), make, 
				methodDesc(XBAFunction.class, Class.class, String.class, String[].class));
	}
	
	
	/* =================================== Expressoes ======================================= */
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTBinaryMessage message) {
		process(message.receiver());
		process(message.argument());
		currMethod().visitInvokeDynamicInsn(message.selector(),
				XBAObject.rraRawType.toMethodDescriptorString(), bootstrap);
		return null;
	}
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTBlock body) {
		Iterator<Node> it = body.expressions().iterator();
		while (it.hasNext()) {
			process(it.next());
			if (it.hasNext()) { // not the last one
				currMethod().visitInsn(POP);
			}
		}
		return null;
	}
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTBooleanLiteral bool) {
		String value = bool.value()? "T" : "F";
		currMethod().visitFieldInsn(GETSTATIC, iname(XBABool.class),
				value, desc(XBAObject.class));
		return null;
	}
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTFunctionApplication funapp) {
		process(funapp.receiver());
		process(funapp.argument());
		currMethod().visitInvokeDynamicInsn("apply",
				XBAObject.rraRawType.toMethodDescriptorString(), bootstrap);
		return null;
	}
		
	protected br.usp.ime.mac499.xba.types.Type process(ASTIdentifier  id) {
		if (bm.lookup(id.name()) == null) {
			throw new RuntimeException("Identificador " + id.name() + " está sem binding");
			//currMethod().visitFieldInsn(GETSTATIC, iname(XBAPrinter.class), 
			//		"print", desc(XBAObject.class));
		} else { 
			bm.lookup(id.name()).generateLoad(this);
		}
		return null;
	}
	protected br.usp.ime.mac499.xba.types.Type process(ASTIf ite) {
		Label ifTrue = new Label();
		Label end = new Label();
		process(ite.condition());
		currMethod().visitFieldInsn(GETSTATIC, iname(XBABool.class),
				"T", desc(XBAObject.class));
		currMethod().visitJumpInsn(IF_ACMPEQ, ifTrue);
		process(ite.elseExpression());
		currMethod().visitJumpInsn(GOTO, end);
		currMethod().visitLabel(ifTrue);		
		process(ite.thenExpression());
		currMethod().visitLabel(end);
		return null;
	}
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTInteger integer) {
		currMethod().visitTypeInsn(NEW, iname(XBAInteger.class));
		currMethod().visitInsn(DUP);
		pushInteger(currMethod(), integer.value());
		currMethod().visitMethodInsn(INVOKESPECIAL, iname(XBAInteger.class),
				"<init>", "(I)V");	
		return null;
	}
	
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTLambda lambda) {return null;}
	
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTLet let) {
		bm.addNewScope();
		for (int k = 0; k < let.numberOfBindings(); k++) {
			process(let.boundExpression(k));
			
			LocalIdentifier lvar = new LocalIdentifier(nLocalsVars++, ti.typeOfLocalVariable());
			bm.addBinding(let.idBinding(k), lvar);
			lvar.generateStore(this);
		}
		process(let.bodyExpression());
		bm.exitCurrScope();
		return null;
	}
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTSelf self) {
		(new LocalIdentifier(0, null)).generateLoad(this);
		return null;
	}
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTString str) {
		currMethod().visitTypeInsn(NEW, iname(XBAString.class));
		currMethod().visitInsn(DUP);
		currMethod().visitLdcInsn(str.value());
		currMethod().visitMethodInsn(INVOKESPECIAL, iname(XBAString.class),
				"<init>", methodDesc(void.class, String.class));
		return null;
	}
	
	protected br.usp.ime.mac499.xba.types.Type process(ASTUnaryMessage message) {
		process(message.receiver());
		currMethod().visitFieldInsn(GETSTATIC, iname(XBAUnit.class), "instance", desc(XBAUnit.class));
		currMethod().visitInvokeDynamicInsn(message.selector(),
				XBAObject.rraRawType.toMethodDescriptorString(), bootstrap);
		return null;
	}
	
	
	
	
	/* ============================================================================ */
	
	public void processChildren(Node node) {
		for (int k = 0; k < node.jjtGetNumChildren(); k++) {
			process(node.jjtGetChild(k));
		}
	}

	public Type process(Node node) {
		MethodType mtype = MethodType.methodType(br.usp.ime.mac499.xba.types.Type.class, node.getClass());
		try {
			MethodHandle mh = MethodHandles.lookup().findVirtual(XBACodeGen.class, "process", mtype);
			return (Type)mh.invokeWithArguments(this,node);
		} catch (NoSuchMethodException e) {
			e.printStackTrace(); 
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (Throwable e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public void process(ASTAssignment assignment) {
	    process(assignment.rvalue());
		bm.lookup(assignment.id()).generateStore(this);
	}
	
	
	private static class FieldVariable extends BindedIdentifier {
		String name;
		String className;
		FieldVariable(String name, String className, br.usp.ime.mac499.xba.types.Type type) {
			super(type);
			this.name = name;
			this.className = className;
		}
		@Override
		public void generateLoad(XBACodeGen codeGen) {
			codeGen.currMethod().visitVarInsn(ALOAD, 0);
			codeGen.currMethod().visitFieldInsn(GETFIELD, className, name, desc(XBAObject.class));
			
		}
		@Override
		public void generateStore(XBACodeGen codeGen) {
			codeGen.currMethod().visitVarInsn(ALOAD, 0);
			codeGen.currMethod().visitInsn(SWAP);
			codeGen.currMethod().visitFieldInsn(PUTFIELD, className, name, desc(XBAObject.class));
		}
	}
	
	
	
	private static class ClassId extends BindedIdentifier {
		String descriptor;
		public ClassId(String className, br.usp.ime.mac499.xba.types.Type type) {
			super(type);
			this.descriptor = "L" + className + ";";
		}
		
		
		
		@Override
		public void generateLoad(XBACodeGen codeGen) {
			// XXX hack para inicializar a classe. Fazer de um jeito mais limpo e menos custoso.
			codeGen.currMethod().visitTypeInsn(NEW, descriptor); 
			codeGen.currMethod().visitInsn(POP);
			
			codeGen.currMethod().visitLdcInsn(Type.getType(descriptor));
			codeGen.currMethod().visitMethodInsn(INVOKESTATIC, iname(XBAObject.class),
					"constructorOf", methodDesc(XBAFunction.class, Class.class));
		}

		@Override
		public void generateStore(XBACodeGen codeGen) {
			throw new RuntimeException("Should not be called");
		}
	}
	
	private static class DefaultInstanceIdentifier extends BindedIdentifier {
		String descriptor;
		String instanceName;
		public DefaultInstanceIdentifier(String className, String instanceName, br.usp.ime.mac499.xba.types.Type type) {
			super(type);
			this.descriptor = "L" + className + ";";
			this.instanceName = instanceName;
			
		}
		@Override
		public void generateLoad(XBACodeGen codeGen) {
			// XXX hack para inicializar a classe. Fazer de um jeito mais limpo e menos custoso.
			codeGen.currMethod().visitTypeInsn(NEW, descriptor); 
			codeGen.currMethod().visitInsn(POP);
			
			codeGen.currMethod().visitLdcInsn(Type.getType(descriptor));
			codeGen.currMethod().visitLdcInsn(instanceName);
			codeGen.currMethod().visitMethodInsn(INVOKESTATIC, iname(XBAObject.class),
					"defaultInstanceOf", methodDesc(XBAObject.class, Class.class, String.class));
		}

		@Override
		public void generateStore(XBACodeGen codeGen) {
			throw new RuntimeException("Should not be called");
		}
	}
	
	private static class LocalIdentifier extends BindedIdentifier {
		int varNumber;
		public LocalIdentifier(int varNumber, br.usp.ime.mac499.xba.types.Type type) {
			super(type);
			this.varNumber = varNumber;
		}
		@Override
		public void generateLoad(XBACodeGen codeGen) {
			codeGen.currMethod().visitVarInsn(ALOAD, varNumber);
		}

		@Override
		public void generateStore(XBACodeGen codeGen) {
			codeGen.currMethod().visitVarInsn(ASTORE, varNumber);
			
		}
	}
	
	// Auxiliar
	private void pushInteger(MethodVisitor mv, int n) {
		if (0 <= n && n <= 5) {
			mv.visitInsn(ICONST_0+ n);
		} else if (Byte.MIN_VALUE <= n && n <= Byte.MAX_VALUE) {
			mv.visitIntInsn(BIPUSH, n);
		} else if (Short.MIN_VALUE <= n && n <= Short.MAX_VALUE) {
			mv.visitIntInsn(SIPUSH, n);
		} else {
			mv.visitLdcInsn(n);
		}		
	}
	
}
