package br.usp.ime.mac499.xba.runtime;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodHandles.Lookup;
import java.lang.invoke.MethodType;
import java.util.Arrays;

public class XBAFunction extends XBAObject {
	private final MethodHandle mh;
	public final String[] selectors;
	
	public XBAFunction(MethodHandle mh, String[] selectors) {
		if (mh.type().parameterCount() != selectors.length)
			throw new IllegalArgumentException();
		this.mh = mh;
		this.selectors = selectors;
	}
	
	@Override
	public XBAFunction method(String selector, Lookup lookup) {
		if (!this.selectors[0].equals(selector))
			throw new XBAObject.MessageNotUnderstood(this.getClass(), selector);
		return this;
	}
	
	public XBAObject apply(XBAObject object) throws Throwable {
		if (mh.type().parameterCount() == 1) { 
			return (XBAObject) mh.invokeExact(object);
		}
		else {
			return new XBAFunction(mh.bindTo(object), tail(selectors));
		}
	}	
	private String[] tail(String[] a) {
		return Arrays.copyOfRange(a, 1, a.length);
	}
	
	
	public static XBAFunction makeXBAFunctionOf(@SuppressWarnings("rawtypes") Class classz, String methodName, String[] selectors, String typeDesc) {
		MethodType type = MethodType.fromMethodDescriptorString(typeDesc, ClassLoader.getSystemClassLoader());
		return makeXBAFunctionOf(classz, methodName, selectors, type, false);
	}
	
	public static XBAFunction makeXBAFunctionOf(@SuppressWarnings("rawtypes") Class classz, String methodName, String[] selectors) {
		return makeXBAFunctionOf(classz, methodName, selectors, XBAObject.rraType(selectors.length), false);
	}
	
	public static XBAFunction makeXBAFunctionOfStatic(@SuppressWarnings("rawtypes") Class classz, String methodName, String[] selectors) {
		return makeXBAFunctionOf(classz, methodName, selectors, XBAObject.rraType(selectors.length), true);
	}
	
	public static XBAFunction makeXBAFunctionOfStatic(@SuppressWarnings("rawtypes") Class classz, String methodName, String[] selectors, String typeDesc) {
		MethodType type = MethodType.fromMethodDescriptorString(typeDesc, ClassLoader.getSystemClassLoader());
		return makeXBAFunctionOf(classz, methodName, selectors, type, true);
	}
	
	public static XBAFunction makeXBAFunctionOf(@SuppressWarnings("rawtypes") Class classz, String methodName, String[] selectors, MethodType type, boolean isStatic) {
		try {
			
			MethodHandle mh = (isStatic)? MethodHandles.lookup().findStatic(classz, methodName, type) :
										  MethodHandles.lookup().findVirtual(classz, methodName, type);
			if (!isStatic) selectors = insertFirstSelector(selectors);
			return new XBAFunction(mh, selectors);
		} catch (NoSuchMethodException | IllegalAccessException e) {
			e.printStackTrace();
			return null;
		}
	}
	
	
	
	
		

	private static String[] insertFirstSelector(String[] selectors) {
		String[] ret = new String[selectors.length + 1];
		ret[0] = "xbaobj";
		for (int k = 0; k < selectors.length; k++) 
			ret[k+1] = selectors[k];
		return ret;
	}
	
	
	// Only called by (java launcher) constructor call?
	public XBAObject applyArgs(XBAObject... args) throws Throwable {
		if (mh.type().parameterCount() != args.length)
			throw new IllegalArgumentException();
		
		return (XBAObject)mh.invokeExact(args); // invokexact?
	}

}
