package br.usp.ime.mac499.xba.transformer;

import static org.objectweb.asm.Opcodes.ASM4;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;


/**
 * Transforma classes JAVA anotadas devidamente em classes XBA válidas. 
 * Ver anotações XBAClass e XBAMethod para maiores informações.
 * @author stagni
 *
 */
public class Transformer {
	
	
	
	public static void main(String[] args) throws IOException {
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES); 
		final XBAInfoReader infoReader = new XBAInfoReader(cw);
		XBAClassAdapter adapter = new XBAClassAdapter(infoReader);
		//ClassReader cr = new ClassReader("br.usp.ime.mac499.xba.transformer.Test");
		ClassReader cr = new ClassReader(args[0]);
		
		cr.accept(adapter, 0);
		
		// Renomeando
		final ClassReader crRen = new ClassReader(cw.toByteArray());
		final ClassWriter cwRen = new ClassWriter(ClassWriter.COMPUTE_FRAMES);
		ClassVisitor cv = new ClassVisitor(ASM4, cwRen) {
			public void visit(int version, int access, String name, String signature, String superName, String[] interfaces) {
				this.cv.visit(version, access, infoReader.className(), signature, superName, interfaces);
			}
		};
		crRen.accept(cv, 0);
		
		FileOutputStream fos = new FileOutputStream(args[1]);
		fos.write(cwRen.toByteArray());
		fos.close();
	}
	
	@org.junit.Test
	public void generateXBAInteger() throws IOException {
		Transformer.main(new String[] { "br.usp.ime.mac499.xba.runtime.XBAInteger", 
				 binPath + "/br/usp/ime/mac499/xba/runtime/XBAInteger.class"}); 
	}
	
	@org.junit.Test
	public void generateXBAString() throws IOException {
		Transformer.main(new String[] { "br.usp.ime.mac499.xba.runtime.XBAString", 
				                        binPath + "/br/usp/ime/mac499/xba/runtime/XBAString.class"}); 
	}
	
	@org.junit.Test
	public void generateXBAPrinter() throws IOException {
		Transformer.main(new String[] { "br.usp.ime.mac499.xba.runtime.XBAPrinter", 
				 binPath + "/br/usp/ime/mac499/xba/runtime/XBAPrinter.class"}); 
	}
	
	private static String binPath;
	static {
		ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
		URL url = classLoader.getResource("");
		try {
			File file = new File(url.toURI());
			binPath = file.toString();
		} catch (URISyntaxException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
}
