package br.usp.ime.mac499.xba.transformer;

import static org.objectweb.asm.Opcodes.ACC_PUBLIC;
import static org.objectweb.asm.Opcodes.ACC_STATIC;
import static org.objectweb.asm.Opcodes.ASM4;
import static org.objectweb.asm.Opcodes.INVOKESTATIC;
import static org.objectweb.asm.Opcodes.INVOKEVIRTUAL;
import static org.objectweb.asm.Opcodes.PUTSTATIC;
import static org.objectweb.asm.Opcodes.RETURN;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Type;


import br.usp.ime.mac499.xba.codegen.XBACodeGen;
import br.usp.ime.mac499.xba.runtime.XBAFunction;
import br.usp.ime.mac499.xba.transformer.XBAInfoReader.XBAMethodInfo;

/** Transforma uma classe Java anotada em uma classe XBA. Necessária para fazer wrappers
 * que acessam classes em Java.
 * 
 * @author stagni
 *
 */
public class XBAClassAdapter extends ClassVisitor  {
	private static final String defaultInitializerName = "xba-static-initializer";
	public final XBAInfoReader extractor;
	private String internalName;
	private boolean hasClassInitializer;
	public XBAClassAdapter(XBAInfoReader cv) {
		super(ASM4, cv);
		this.extractor = cv;
		this.hasClassInitializer = false;
	}

	@Override
	public void visit(int version, int access, String name, String signature, String superName, String[] interfaces) {
		this.internalName = name;
		cv.visit(version, access, name, signature, superName, interfaces);
	}
	
	@Override 
	public MethodVisitor visitMethod(int access, String name, String desc, String signature, String[] exceptions) {
		MethodVisitor mv = cv.visitMethod(access, name, desc, signature, exceptions);
		if (name.equals("<clinit>") && desc.equals("()V")) {
			hasClassInitializer = true;
			return new MethodVisitor(ASM4, mv) {
				@Override
				public void visitInsn(int opcode) {   
					if (opcode == RETURN) { 
						mv.visitMethodInsn(INVOKESTATIC, internalName, defaultInitializerName, "()V");
						mv.visitInsn(RETURN);
					} else {
						mv.visitInsn(opcode);
					}
				}
			};
		}
		return mv;
	}
	
	@Override
	public void visitEnd() {
		String initializerName = (hasClassInitializer)? defaultInitializerName : "<clinit>";
		MethodVisitor mv = cv.visitMethod(ACC_STATIC, initializerName, "()V", null, null);
		String makedescriptor = XBACodeGen.methodDesc(XBAFunction.class, Class.class, String.class, String[].class, String.class);
		Type classType = Type.getType("L" + internalName + ";");
		
		mv.visitCode();
		for (XBAMethodInfo methodInfo : extractor.methodsInfo()) {
			String functionName = (methodInfo.isStatic)? "makeXBAFunctionOfStatic" : "makeXBAFunctionOf";
			String fieldName = (methodInfo.isStatic)? "$constructor" : methodInfo.firstSelector();
			
								
			// Cria o field que guardara a function
			cv.visitField(ACC_PUBLIC + ACC_STATIC, fieldName,
					XBACodeGen.desc(XBAFunction.class), null, null).visitEnd();
			// A classe que está sendo gerada..
			mv.visitLdcInsn(classType);
			// o nome do método 
			mv.visitLdcInsn(methodInfo.name);
			// TODO: preguiça de criar a array elemento por elemento, entao usando split....:
			mv.visitLdcInsn(methodInfo.selectorDescriptor());
			mv.visitLdcInsn(":");
			mv.visitMethodInsn(INVOKEVIRTUAL, XBACodeGen.iname(String.class), 
					"split", XBACodeGen.methodDesc(String[].class, String.class));
			// ---------------------------------------------------------------- 
			mv.visitLdcInsn(methodInfo.descriptor);
			mv.visitMethodInsn(INVOKESTATIC, XBACodeGen.iname(XBAFunction.class), 
					functionName, makedescriptor);
			mv.visitFieldInsn(PUTSTATIC, internalName, fieldName, XBACodeGen.desc(XBAFunction.class));
		}
		mv.visitInsn(RETURN);
		mv.visitMaxs(-1, -1);
		mv.visitEnd();
		cv.visitEnd();
	}
	
		
}
