package br.usp.ime.mac499.xba.types;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/** 
 * Representa o tipo de um objeto
 * @author stagni
 *
 */
public class ObjectType extends Type {
	private Map<String, Arrow> selector2arrow;
	private final boolean isClosed;
	public ObjectType() {
		this(true);
	}
	
	public boolean isClosed() {
		return isClosed; 
	}
	
	public ObjectType(boolean closed) {
		this.isClosed = closed;
		this.selector2arrow = new HashMap<String, Arrow>();
	}
	
	public void add(String selector, Arrow arrow) {
		selector2arrow.put(selector, arrow);
	}
	
	public void add(String selector, Type from, Type to) {
		selector2arrow.put(selector, new Arrow(from, to));
	}
	
	public Arrow get(String selector) {
		return selector2arrow.get(selector);
	}
	
	public boolean hasSelector(String selector) {
		return selector2arrow.containsKey(selector);
	}
	
	public Set<String> selectors() {
		return selector2arrow.keySet();
	}
	
	public static ObjectType singleArrowObject(String selector, Type from, Type to) {
		return singleArrowObject(selector, from , to, true);
	}
	public static ObjectType singleArrowObject(String selector, Type from, Type to, boolean closed) {
		ObjectType ret = new ObjectType(closed);
		ret.add(selector, from, to);
		return ret;	
	}
	
}


