package br.usp.ime.mac499.xba.types;

import static org.objectweb.asm.Opcodes.ASM4;

import java.io.IOException;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;

import br.usp.ime.mac499.xba.transformer.XBAClass;

/**
 * Representa um tipo que está apenas referenciando um tipo previamente serializado.
 * @author stagni
 *
 */
public class RefType extends Type {
	public final String qualifedClassName;
	/**
	 * Cria um novo tipo que é a referência para um tipo previamente serializado em uma 
	 * classe. Esse será o tipo DA CLASSE e não das instâncias.
	 * @param qualifedClassName nome da classe 
	 */
	public RefType(String qualifedClassName) {
		this.qualifedClassName = qualifedClassName;
	}
	
	transient Type cachedReferencedType = null;
	
	@Override
	public Type realType() {
		if (cachedReferencedType != null) 
			return cachedReferencedType; 
		try {
			cachedReferencedType = referenceType();
		} catch (IOException e) {} // FIXME: swallow.
		return cachedReferencedType;
	}

	private Type referenceType() throws IOException {
		ClassReader cr = new ClassReader(qualifedClassName);
		TypeClassExtractor tcr = new TypeClassExtractor();
		cr.accept(tcr, 0);
		TypeDeserializer td = new TypeDeserializer();
		return td.deserializeFromString(tcr.serializedType());
	}
	
	private static class TypeClassExtractor extends ClassVisitor {
		private String serializedType = null;
		public TypeClassExtractor() {
			super(ASM4);
		}
		
		public String serializedType() {
			if (serializedType == null) throw new RuntimeException("Tipo serializado é inválido.");
			return serializedType;
		}
		@Override 
		public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
			AnnotationVisitor av = cv.visitAnnotation(desc, visible);
			if (desc.equals(org.objectweb.asm.Type.getDescriptor(XBAClass.class))) {
				return new AnnotationVisitor(ASM4, av) {
					@Override
					public void visit(String name, Object value) {
						if (name.equals("type") && value instanceof String) 
							serializedType = (String)value; 
						av.visit(name, value);
					}
				};
			}
			return av;
		}
	}
	

}
