package br.usp.ime.mac499.xba.types;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

public class TypeDeserializer {
	protected Map<Integer, Type> id2type = new HashMap<Integer, Type>();
	// TODO: colocar em classe a parte
	public static final char REGISTERED = 'A';
	public static final char OBJECT = 'O';
	public static final char OPEN = 'O';
	public static final char CLOSED = 'C';
	public static final char VARIABLE = 'V';
	public static final char REF = 'R';
	public static final char END = ';';
	
	public Type deserializeFromString(String string) throws IOException {
		ByteArrayInputStream bais = new ByteArrayInputStream(string.getBytes());
		return deserialize(bais);
	}
	
	
	public Type deserialize(InputStream stream) throws IOException {
		char firstChar = (char)stream.read();
		return deserialize(stream, firstChar);
	}
	
	public Type deserialize(InputStream stream, char firstChar) throws IOException {
		if (firstChar == 'A')
			return deserializeRegistered(stream);
		else if (firstChar == 'O') 
			return deserializeObject(stream);
		else if (firstChar == 'V')
			return deserializeVariable(stream);
		else if (firstChar == 'R')
			return deserializeRef(stream);
		else
			throw new RuntimeException("invalid stream: " + firstChar);

			
	}
	
	public Type deserializeRegistered(InputStream stream) throws IOException {
		int id = readInt(stream);
		char end = (char)stream.read(); assert(end == END);
		return id2type.get(id);
	}
	
	public TypeVariable deserializeVariable(InputStream stream) throws IOException {
		int id = readInt(stream);
		TypeVariable var = new TypeVariable();
		id2type.put(id, var);
		char end = (char)stream.read(); assert(end == END);
		return var;
	}
	
	public ObjectType deserializeObject(InputStream stream) throws IOException {
		int id = readInt(stream);
		int closed = stream.read();
		ObjectType ret = new ObjectType(closed == CLOSED);
		id2type.put(id, ret);
		
		char read;
		while((read = (char)stream.read()) != END) { // FIXME: checar se devolve -1  
			Type from = deserialize(stream, read);
			String selector = readString(stream);
			Type to = deserialize(stream);
			ret.add(selector, from, to);
		}
		char end = (char)stream.read(); assert(end == END);
		return ret;
	}
	
	public RefType deserializeRef(InputStream stream) throws IOException {
		int id = readInt(stream);
		String className = readString(stream);
		RefType ret = new RefType(className);
		id2type.put(id, ret);
		char end = (char)stream.read(); assert(end == END);
		return ret;
	}
	
	private int readInt(InputStream stream) throws IOException {
		String ret = "";
		char read;
		while((read = (char)stream.read()) != END) // FIXME: checar se devolve -1  
			ret += read;
		return Integer.parseInt(ret);
	
	}
	
	private String readString(InputStream stream) throws IOException {
		String ret = "";
		char read;
		while((read = (char)stream.read()) != END) // FIXME: checar se devolve -1  
			ret += read;
		return ret;
	}
	
	
	
}
