package br.usp.ime.mac499.xba.types.inference;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import br.usp.ime.mac499.xba.ast.ASTBinaryMessage;
import br.usp.ime.mac499.xba.ast.ASTBlock;
import br.usp.ime.mac499.xba.ast.ASTClass;
import br.usp.ime.mac499.xba.ast.ASTFieldDefinition;
import br.usp.ime.mac499.xba.ast.ASTIdentifier;
import br.usp.ime.mac499.xba.ast.ASTInteger;
import br.usp.ime.mac499.xba.ast.ASTLet;
import br.usp.ime.mac499.xba.ast.ASTMethodDefinition;
import br.usp.ime.mac499.xba.ast.ASTSelf;
import br.usp.ime.mac499.xba.ast.Node;
import br.usp.ime.mac499.xba.types.BasicTypes;
import br.usp.ime.mac499.xba.types.ObjectType;
import br.usp.ime.mac499.xba.types.Type;
import br.usp.ime.mac499.xba.types.Type.Arrow;
import br.usp.ime.mac499.xba.types.TypeVariable;
import br.usp.ime.mac499.xba.util.BindingsManager;

/**
 * Classe de inferência de tipos provisória. Deve ser integrada a XBACodeGen (ver 
 * TypeInferer para início da implementação dessa integração)
 * @author stagni
 *
 */
public class TypeChecker {
	BindingsManager<Type> bm;
	Unifier unifier;
	public TypeChecker() {
		unifier = new Unifier();
		bm = new BindingsManager<Type>();
	}
	
	public Type typecheck(Node node) {
		MethodType mtype = MethodType.methodType(Type.class, node.getClass());
		try {
			MethodHandle mh = MethodHandles.lookup().findVirtual(this.getClass(), "typecheck", mtype);
			return (Type)mh.invokeWithArguments(this,node);
		} catch (NoSuchMethodException e) {
			e.printStackTrace(); 
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (TypeException e) {
			throw e;
		} catch (Throwable e) {
			e.printStackTrace();
		}
		throw new RuntimeException("visitor dispatch error");
	}
	
	public Type typecheck(ASTClass classz) {
		ObjectType classType = (ObjectType)buildSelectorsMethodTypeRec(classz, 0); //FIXME: CAST
		ObjectType self = new ObjectType();
		bm.addNewScope();
		bm.addBinding("self", self);
		bm.addBinding(classz.name(), classType);
		// Monta estrutura do objeto 
		// Fields.. e adiciona bindings para fields
		for (ASTFieldDefinition field : classz.fields()) 
			bm.addBinding(field.name(), new TypeVariable());
		for (ASTMethodDefinition method : classz.methods()) {
			Type t = buildSelectorsMethodTypeRec(method, 1);
			self.add(method.selector(0), new TypeVariable(), t);
		}
		for (ASTMethodDefinition method : classz.methods()) 
			typecheck(method);
		
		Type ret = unifier.extract(typecheckConstructor(classz, classType, self));
		bm.exitCurrScope();		
		return ret;
	}
	
	
	
	private Type typecheckConstructor(ASTClass classz, ObjectType classType, ObjectType self) {
		Arrow arrow = classType.get(classz.selector(0));
		bm.addNewScope();		
		bm.addBinding(classz.parameter(0), arrow.from);
		for (int k = 1; k < classz.numberOfParameters(); k++) {
			arrow = ((ObjectType)arrow.to).get(classz.selector(k));
			bm.addBinding(classz.parameter(k), arrow.from);
		}
		
		for (ASTFieldDefinition f : classz.fields()) {
			Type fieldType = typecheck(f.initializedTo());
			if (!unifier.unify(fieldType, bm.lookup(f.name()))) //FIXME: e se tiver um parametro com mesmo nome do campo??
				return null;
		}
		if (!unifier.unify(self, arrow.to))
			return null;
			
		bm.exitCurrScope();
		return classType;
	}
	
	private Type buildSelectorsMethodTypeRec(ASTMethodDefinition method, int i) {
		if (i >= method.numberOfParameters())
			return new TypeVariable();
		Type to = buildSelectorsMethodTypeRec(method, i+1);
		return ObjectType.singleArrowObject(method.selector(i), new TypeVariable(), to);
	}
	
	
	
	public boolean typecheck(ASTMethodDefinition method) {
		ObjectType self = (ObjectType)bm.lookup("self");
		Arrow arrow = self.get(method.selector(0));
		// Adiciona parâmetros ao escopo
		bm.addNewScope();
		bm.addBinding(method.parameter(0), arrow.from);
		for (int k = 1; k < method.numberOfParameters(); k++) {
			arrow = ((ObjectType)arrow.to).get(method.selector(k));
			bm.addBinding(method.parameter(k), arrow.from);
		}
		Type ret = typecheck(method.body());
		if (!unifier.unify(arrow.to, ret))
			return false;
		bm.exitCurrScope();
		return true;
	}
	
	public Type typecheck(ASTBinaryMessage message) {
		Type receiverType = typecheck(message.receiver());
		Type argumentType = typecheck(message.argument());
		Type parameterType = new TypeVariable();
		Type returnType = new TypeVariable();
		ObjectType rcvExpected = ObjectType.singleArrowObject(message.selector(), 
				parameterType, returnType, false);
		if (unifier.unify(rcvExpected, receiverType) == false)
			return null; // FIXME throw exception
		if (unifier.unify(parameterType, argumentType) == false)
			return null; // FIXME throw exception
		return unifier.extract(returnType);
	}
	
	public Type typecheck(ASTLet let) {
		bm.addNewScope();
		Type boundType = typecheck(let.boundExpression(0));
		bm.addBinding(let.idBinding(0), boundType);
		Type ret = typecheck(let.bodyExpression());
		bm.exitCurrScope();
		return ret;
	}
	
	public Type typecheck(ASTInteger i) {
		return BasicTypes.Int;
	}
	
	public Type typecheck(ASTBlock block) {
		Node last = block.last();
		for (Node node : block.expressions()) {
			if (node != last)
				typecheck(node);
		}
		return typecheck(last);
	}
	
	public Type typecheck(ASTSelf self) {
		return bm.lookup("self");
	}
	
	public Type typecheck(ASTIdentifier id) {
		Type ret = bm.lookup(id.name());
		if (ret == null)
			return null; // FIXME: throw exception
		return ret;
	}
	
	public static class TypeException extends RuntimeException {
		private static final long serialVersionUID = 2814218670013582773L;
				
	}
}
