package br.usp.ime.mac499.xba.types.inference;

import br.usp.ime.mac499.xba.ast.ASTBinaryMessage;
import br.usp.ime.mac499.xba.ast.ASTBooleanLiteral;
import br.usp.ime.mac499.xba.ast.ASTClass;
import br.usp.ime.mac499.xba.ast.ASTFieldDefinition;
import br.usp.ime.mac499.xba.ast.ASTInteger;
import br.usp.ime.mac499.xba.ast.ASTMethodDefinition;
import br.usp.ime.mac499.xba.ast.ASTString;
import br.usp.ime.mac499.xba.ast.ASTUnaryMessage;
import br.usp.ime.mac499.xba.codegen.BindedIdentifier;
import br.usp.ime.mac499.xba.types.ObjectType;
import br.usp.ime.mac499.xba.types.RefType;
import br.usp.ime.mac499.xba.types.Type;
import br.usp.ime.mac499.xba.types.Type.Arrow;
import br.usp.ime.mac499.xba.types.TypeVariable;
import br.usp.ime.mac499.xba.util.BindingsManager;

/**
 * Classe que não está sendo usada atualmente pelos testes, mas deverá 
 * ser usada quando a inferência de tipos estiver integrada com a geração de 
 * código.
 * 
 * @author stagni
 */
public class TypeInferer {
	
	private static final Type unit = new RefType("br.usp.ime.mac499.xba.runtime.XBAUnit");
	private final BindingsManager<BindedIdentifier> bm;
	private final Unifier unifier;
	public TypeInferer(BindingsManager<BindedIdentifier> bm) {
		this.bm = bm;
		this.unifier = new Unifier();
	}
	
	public Type typeOfInteger(ASTInteger integer) {
		return new RefType("br.usp.ime.mac499.xba.runtime.XBAInteger");
	}
	
	public Type typeOfString(ASTString string) {
		return new RefType("br.usp.ime.mac499.xba.runtime.XBAInteger");
	}
	
	public Type typeOfBoolean(ASTBooleanLiteral bool) {
		return new RefType("br.usp.ime.mac499.xba.runtime.XBABool");
	}
	
	public Type typeOfClassId(String classId) {
		return new RefType(classId);
	}
	
	public Type typeOfBinMessage(ASTBinaryMessage message, Type receiverType, Type argumentType) {		
		return typeOfMessage(message.selector(), receiverType, argumentType);
	}
	
	public Type typeOfUnaryMessage(ASTUnaryMessage unary, Type receiverType) {
		return typeOfMessage(unary.selector(), receiverType, unit);
	}
	
	public Type typeOfField(ASTFieldDefinition field) {
		return new TypeVariable();
	}
	
	public Type typeOfLocalVariable() {
		return new TypeVariable();
	}
	
	public Type typeOfClassInitial(ASTClass classz) {
		return buildSelectorsMethodTypeRec(classz, 0);
	}
	
	public Type typeOfSelf(ASTClass classz) {
		ObjectType self = new ObjectType();
		for (ASTMethodDefinition method : classz.methods()) {
			Type t = buildSelectorsMethodTypeRec(method, 1);
			self.add(method.selector(0), new TypeVariable(), t);
		}
		return self;
	}
	
	public Type typeOfConstructor(ASTClass classz) {
		return null;
	}
	
	public Type typeOfMethodArgument(ASTMethodDefinition methodDef, int k) {
		return null;
	}
	
	private Type typeOfMessage(String selector, Type receiverType, Type argumentType) {		
		Type parameterType = new TypeVariable();
		Type returnType = new TypeVariable();
		ObjectType rcvExpected = ObjectType.singleArrowObject(selector, 
				parameterType, returnType, false);
		if (unifier.unify(rcvExpected, receiverType) == false)
			throw new TypeInferenceException();
		if (unifier.unify(parameterType, argumentType) == false)
			throw new TypeInferenceException();
		return unifier.extract(returnType);
	}
	
	
	private Type buildSelectorsMethodTypeRec(ASTMethodDefinition method, int i) {
		if (i >= method.numberOfParameters())
			return new TypeVariable();
		Type to = buildSelectorsMethodTypeRec(method, i+1);
		return ObjectType.singleArrowObject(method.selector(i), new TypeVariable(), to);
	}
	
	public static class TypeInferenceException extends RuntimeException {
		private static final long serialVersionUID = 2814218660013582773L;
				
	}
}
