package br.usp.ime.mac499.xba.types.inference;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.util.HashMap;
import java.util.Map;

import br.usp.ime.mac499.xba.types.ObjectType;
import br.usp.ime.mac499.xba.types.Type;
import br.usp.ime.mac499.xba.types.Type.Arrow;
import br.usp.ime.mac499.xba.types.TypeVariable;

/**
 * Classe que tenta dar um feedback visual a tipos mais legível do que a serialização.
 * @author stagni
 *
 */
public class TypePrinter {
	private Character special = new Character('?');
	private Map<ObjectType, Character> type2symbol = new HashMap<ObjectType, Character>();
	private Map<TypeVariable, Character> var2symbol = new HashMap<TypeVariable, Character>();
	private char nextTypeChar = 't';
	private char nextVarChar = 'a';
	
	
	private char nextAvaliableTypeChar() {
		char ret = nextTypeChar;
		nextTypeChar = (char)(((int)nextTypeChar) + 1);
		return ret;
	}
	private char nextAvaliableVarChar() {
		char ret = nextVarChar;
		nextVarChar = (char)(((int)nextVarChar) + 1);
		return ret;
	}
	
	private String printTT(Type t) {
		try {
			MethodType mtype = MethodType.methodType(String.class, t.getClass());
			MethodHandle mh = MethodHandles.lookup().findVirtual(this.getClass(), "printT", mtype);		
			return (String)mh.bindTo(this).invokeWithArguments(t);
		} catch (NoSuchMethodException e) {		
		} catch (IllegalAccessException e) {
		} catch (Throwable e) {
			e.printStackTrace();
		}
		throw new RuntimeException("DD FAIL");
	}
	
	protected String printT(ObjectType t) {
		Character symbol = type2symbol.get(t);
		if (symbol == null) {
			String ret = "{";
			type2symbol.put(t, special);
			for (String selector : t.selectors()) {
				Arrow arrow = t.get(selector);
				ret += printTT(arrow.from) + "-(" + selector + ")->" + printTT(arrow.to) + " ; ";  
			}
			if (type2symbol.get(t) != special) { // tipo recursivo
				ret = "u" + type2symbol.get(t).toString() + ret;
			}
			return ret + ((t.isClosed())? "":"...") + "}";
		}
		else if (symbol == special) 
			type2symbol.put(t, nextAvaliableTypeChar());
		
		return type2symbol.get(t).toString();
	}
	
	protected String printT(TypeVariable t) {
		if (var2symbol.get(t) == null)
			var2symbol.put(t, nextAvaliableVarChar());
		return var2symbol.get(t).toString();
	}
	
	public static String print(Type t) {
		return (new TypePrinter()).printTT(t);
	}
}
