package br.usp.ime.mac499.xba.types.inference;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import br.usp.ime.mac499.xba.types.ObjectType;
import br.usp.ime.mac499.xba.types.Type;
import br.usp.ime.mac499.xba.types.TypeVariable;
import br.usp.ime.mac499.xba.types.Type.Arrow;

public class Unifier {
	private Map<Type, Type> type2parent;
	public Unifier() {
		type2parent = new HashMap<Type, Type>();
	}
	
	private Type parent(Type t) {
		Type parent;
		for (parent = t; type2parent.get(parent) != null; parent = type2parent.get(parent))
			; // TODO: Fazer COMPACTAÇÃO do union-find. (É só uma linha, mas é melhor fazer sem otimização por enqnto) 
		return parent;
	}
	
	private void assignParent(Type t, Type parent) {
		type2parent.put(t, parent);
	}
	// Double dispatch (comutativo)
	public boolean unify(Type type1, Type type2) {
		Type par1 = parent(type1).realType();
		Type par2 = parent(type2).realType();
		if (par1 == par2) 
			return true;
		try {
			MethodType mtype = MethodType.methodType(boolean.class, par1.getClass(), par2.getClass());
			MethodHandle mh = MethodHandles.lookup().findVirtual(this.getClass(), "unify", mtype);		
			return (Boolean)mh.bindTo(this).invokeWithArguments(par1, par2);
		} catch (NoSuchMethodException e) {
			return unify(par2, par1); // Comutativo! XXX: stackoverflow se não colocar todas as possibilidades!!
		} catch (IllegalAccessException e) {
		} catch (Throwable e) {
		}
		throw new RuntimeException("DD FAIL");
	}
	
	protected boolean unify(TypeVariable t1, TypeVariable t2) {
		assignParent(t1, t2);
		return true;
	}
	
	protected boolean unify(TypeVariable t1, ObjectType t2) {
		assignParent(t1, t2); // variavel NUNCA pode ser pai de nao-variavel
		return true;
	}
	
	protected boolean unify(ObjectType t1, ObjectType t2) {
		// TODO: DESNECESSÁRIO completar ambos os tipos: é perda de tempo pois só 1 será o parent...
		// Da pra fazer melhor, chamando funcoes diferentes no caso aberto/fechado, aberto/aberto..
		boolean ret = unifyBigSmallObj(t1, t2) & unifyBigSmallObj(t2, t1);
		if (t1.isClosed()) 
			assignParent(t2, t1); // Se ambos forem fechados tanto faz quem é o pai.
		else 
			assignParent(t1, t2); // Se ambos forem abertos tanto faz quem é o pai.
		return ret;
		
	}
	
	
	private boolean unifyBigSmallObj(ObjectType big, ObjectType small) {
		for (String selector : big.selectors()) {
			if (!small.hasSelector(selector)) {
				if (small.isClosed()) return false;
				small.add(selector, new TypeVariable(), new TypeVariable());
			}
			Arrow bigArrow = big.get(selector);
			Arrow smallArrow = small.get(selector);
			if (!unify(bigArrow.from, smallArrow.from)) return false;
			if (!unify(bigArrow.to, smallArrow.to)) return false;
		}
		return true;
	}
	
	public Type extract(Type t) {
		return (new Cleaner()).cleanUp(t);
	}
	
	private class Cleaner {
		Set<Type> cleaned = new HashSet<Type>();
		
		public Type cleanUp(Type t) {
			Type parent = parent(t);
			if (cleaned.contains(parent)) return parent;
			cleaned.add(parent);
			if (parent instanceof ObjectType) {
				ObjectType obj = (ObjectType)parent;
				for (String selector : obj.selectors()) {
					Arrow arrow = obj.get(selector);
					arrow.to = cleanUp(arrow.to);
					arrow.from = cleanUp(arrow.from);
				}
			}
			return parent;
		}
		
	}
	
	
	
}
