package br.usp.ime.mac499.xba.parser;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;

import org.junit.Test;

import br.usp.ime.mac499.xba.ast.ASTClass;
import br.usp.ime.mac499.xba.ast.ASTModule;
import br.usp.ime.mac499.xba.ast.Node;

/**
 * Diversos testes verificando que o parser constroi árvores sintáticas corretas 
 * para classes
 * @author stagni
 *
 */
public class ClassParserTest {
	
	@Test
	public void canParseSimpleClassOnlyWithMembers() throws UnsupportedEncodingException, FileNotFoundException, ParseException {
		XBAFullParser parser = new XBAFullParser("progs/incomplete/class/simple-class-with-members-only");
		Node node = parser.XBAClass();
		assertTrue(node instanceof ASTModule);
		ASTClass c = (ASTClass) node.jjtGetChild(0);
		assertEquals("Simple", c.name());
		assertEquals(3, c.fields().size());
		assertEquals(0, c.methods().size());
		assertEquals(0, c.exports().size());
	}
	
	@Test
	public void canParseSimpleClassOnlyWithMethods() throws UnsupportedEncodingException, FileNotFoundException, ParseException {
		XBAFullParser parser = new XBAFullParser("progs/incomplete/class/simple-class-with-methods-only");
		Node node = parser.XBAClass();
		assertTrue(node instanceof ASTModule);
		ASTClass c = (ASTClass) node.jjtGetChild(0);
		assertEquals("Simple", c.name());
	}
		
	
	@Test
	public void canParseSimpleClass() throws UnsupportedEncodingException, FileNotFoundException, ParseException {
		XBAFullParser parser = new XBAFullParser("progs/incomplete/class/simple-class");
		Node node = parser.XBAClass();
		assertTrue(node instanceof ASTModule);
		ASTClass c = (ASTClass) node.jjtGetChild(0);
		assertEquals("Simple", c.name());
	}
	
	@Test
	public void canParseSimpleClass2() throws UnsupportedEncodingException, FileNotFoundException, ParseException {
		XBAFullParser parser = new XBAFullParser("progs/incomplete/class/printer");
		Node node = parser.XBAClass();
		assertTrue(node instanceof ASTModule);
		ASTClass c = (ASTClass) node.jjtGetChild(0);
		assertEquals("Printer", c.name());
		
	}
	
	@Test
	public void canParseHelloWorld() throws UnsupportedEncodingException, FileNotFoundException, ParseException {
		XBAFullParser parser = new XBAFullParser("progs/incomplete/class/helloworld");
		Node node = parser.XBAClass();
		assertTrue(node instanceof ASTModule);
		ASTClass c = (ASTClass) node.jjtGetChild(0);
		assertEquals("Hello", c.name());
		
	}
	
	@Test
	public void canParseMdc() throws UnsupportedEncodingException, FileNotFoundException, ParseException {
		XBAFullParser parser = new XBAFullParser("progs/incomplete/class/mdc");
		Node node = parser.XBAClass();
		assertTrue(node instanceof ASTModule);
		ASTClass c = (ASTClass) node.jjtGetChild(0);
		assertEquals("Mdc", c.name());
		assertEquals(0, c.fields().size());
		assertEquals(1, c.methods().size());
		assertEquals(1, c.exports().size());
	}
	
	@Test
	public void canParseVector() throws UnsupportedEncodingException, FileNotFoundException, ParseException {
		XBAFullParser parser = new XBAFullParser("progs/incomplete/class/vector");
		Node node = parser.XBAClass();
		assertTrue(node instanceof ASTModule);
		ASTClass c = (ASTClass) node.jjtGetChild(0);
		assertEquals("Vector", c.name());
		
	}
}
