package br.usp.ime.mac499.xba.preprocesser;

import static org.junit.Assert.assertEquals;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;

import org.junit.Test;

import br.usp.ime.mac499.xba.preprocess.WhitespacePreprocesserReader;

public class WhitespacePreprocesserReaderTest {
	WhitespacePreprocesserReader whiteReader;
	
	private final String line = "uma    linha    de   codigo";
	
	
	private String fromReader(Reader reader) throws IOException {
		StringBuilder sb = new StringBuilder();
		int nextChar;
		while ((nextChar = reader.read()) != -1)
			sb.append((char)nextChar);
		return sb.toString();
	}
	
	
	public void readFile(String fileName) throws IOException {
		System.out.println(fromReader(new WhitespacePreprocesserReader(new InputStreamReader(new FileInputStream(fileName), "UTF-8"))));
	}
	
	@Test
	public void canReadFile() throws IOException
	{
		readFile("progs/incomplete/block/p1.xba");
		
	}
	@Test
	public void canReadNonWhiteString() throws IOException {
		String nonWhite = "nonWhite\n";
		Reader reader = new WhitespacePreprocesserReader(new StringReader(nonWhite));
		assertEquals(nonWhite, fromReader(reader));		
	}
	
	@Test (expected = WhitespacePreprocesserReader.InvalidCharacterException.class)
	public void throwsExceptionWhenReadingTabs() throws IOException {
		String nonWhite = "\tnonWhite\n";
		Reader reader = new WhitespacePreprocesserReader(new StringReader(nonWhite));
		fromReader(reader);				
	}
	
	@Test
	public void canGenerateIndentCharactersInSimpleExample() throws IOException {
		String str = line + "\n" +
					 "    " + line + "\n";	
		
		WhitespacePreprocesserReader reader = new WhitespacePreprocesserReader(new StringReader(str));
		assertEquals(line + reader.indentChar() + "\n" + "    " +  line + reader.dedentChar() + "\n", fromReader(reader));			
	}
	
	@Test(expected = WhitespacePreprocesserReader.BadDedentException.class)
	public void throwsExceptionOnBadDedent() throws IOException {
		String str = line + "\n" +
					 "    " + line + "\n" +
		             "  " + line;
		
		WhitespacePreprocesserReader reader = new WhitespacePreprocesserReader(new StringReader(str));
		fromReader(reader);			
	}
	
	@Test
	public void canGenerateIndentCharactersInExampleNotEndingWithEmptyLine() throws IOException {
		String str = line + "\n" +
				 "    " + line;	
	
		WhitespacePreprocesserReader reader = new WhitespacePreprocesserReader(new StringReader(str));
		assertEquals((line + reader.indentChar() + "\n" + "    " +  line + reader.dedentChar() + "\n"), fromReader(reader));
	}
	

	@Test
	public void canGenerateIndentCharactersInComplexExample() throws IOException {
		String str = line + "\n" +
				     "    " + line + "\n" +
				     "    " + line + "\n" +
				     "    " + "     " + line +"\n" +
				     "    " + "     " + "     " + line +"\n" +
				     "    " + line + "\n"; 
				     
		WhitespacePreprocesserReader reader = new WhitespacePreprocesserReader(new StringReader(str));
		String expected = line + reader.indentChar() + "\n" +
			     		  "    " + line + "\n" +
			     		  "    " + line + reader.indentChar() + "\n" +
			     		  "    " + "     " + line + reader.indentChar() + "\n" +
			     		  "    " + "     " + "     " + line + reader.dedentChar() + reader.dedentChar() + "\n" +
			     		  "    " + line + reader.dedentChar() + "\n"; 
		String actual = fromReader(reader);
		assertEquals(expected, actual);			
	}
	
	@Test
	public void genFromFile() throws IOException {
		WhitespacePreprocesserReader reader = new WhitespacePreprocesserReader(new InputStreamReader(new FileInputStream("progs/incomplete/class/vector"), "UTF-8"));
		int c;
		while((c = reader.read()) != -1) {
			System.out.print((char)c);
		}
	}
		
}
