package br.usp.ime.mac499.xba.types;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.IOException;

import org.junit.Test;

public class SerializationTest {
	String expectedSimple = "O0;CV1;;sel1;V2;;A1;;sel3;A0;;A2;;sel2;A1;;;;";
	String expectedComplex = "O0;OR1;br.usp.ime.mac499.xba.Integer;;sel1;A1;;O2;C;;sel2;V3;;;;";
	@Test
	public void canSerializeSimpleType() {
		TypeVariable var1 = new TypeVariable();
		TypeVariable var2 = new TypeVariable();
		ObjectType obj = new ObjectType();
		obj.add("sel1", var1, var2);
		obj.add("sel2", var2, var1);
		obj.add("sel3", var1, obj);
		
		TypeSerializer ts = new TypeSerializer();
		assertEquals(expectedSimple, ts.serializeToString(obj));
	}
	
	@Test
	public void canDeserializeSimpleType() throws IOException {
		TypeDeserializer td = new TypeDeserializer();
		Type t = td.deserializeFromString(expectedSimple);
		assertTrue(t instanceof ObjectType);
		ObjectType obj = (ObjectType) t;
		assertTrue(obj.get("sel1") != null);
		assertTrue(obj.get("sel2") != null);
		assertTrue(obj.get("sel3") != null);
		
		assertEquals(obj, obj.get("sel3").to);
	}
	
	@Test 
	public void canSerializeComplexType() {
		RefType ref = new RefType("br.usp.ime.mac499.xba.Integer");
		ObjectType obj = new ObjectType(false);
		obj.add("sel1", ref, ref);
		obj.add("sel2", new ObjectType(), new TypeVariable());
		
		TypeSerializer ts = new TypeSerializer();
		assertEquals(expectedComplex, ts.serializeToString(obj));
	}
	
	@Test 
	public void canDeserializeComplexType() throws IOException {
		TypeDeserializer ts = new TypeDeserializer();
		Type t = ts.deserializeFromString(expectedComplex);
		
	}
	
	
	
}
